#include <linux/mm.h>
#include <linux/in.h>
#include <linux/locks.h>

#include "spbm.h"
#include "thread.h"

#define normalizetv(t) (t)->tv_sec+=(t)->tv_usec/1000000; (t)->tv_usec=(t)->tv_usec%1000000;

static int pid;
static atomic_t kill_thread;
static atomic_t is_dead;
static wait_queue_head_t my_wait;
static struct task_struct *spbmthread = 0;
static LIST_HEAD(events);
static rwlock_t event_lock = RW_LOCK_UNLOCKED;

struct spbm_event beaconevt;

struct spbm_event *event_find_action(void (*action) (struct spbm_event *))
{
	struct list_head *n;
	unsigned long flags;
	read_lock_irqsave(&event_lock, flags);
	list_for_each(n, &events) {
		if (((struct spbm_event *) n)->action == action)
			return (struct spbm_event *) n;
	}
	read_unlock_irqrestore(&event_lock, flags);
	return 0;
}

void event_changed()
{
	if (current != spbmthread) {
//              wake_up(&my_wait);
		if (spbmthread)
			wake_up_process(spbmthread);
	}
}

void event_add(struct spbm_event *evt)
{
	unsigned long flags;
	write_lock_irqsave(&event_lock, flags);
	list_add((struct list_head *) evt, &events);
	write_unlock_irqrestore(&event_lock, flags);
	event_changed();
}

void event_set_time(struct spbm_event *evt, int microsecs)
{
	do_gettimeofday(&evt->t);
	evt->t.tv_usec += microsecs * 1000;
	normalizetv(&evt->t);
}

void killme()
{
	int count = 0;
	wait_queue_head_t queue;
	init_waitqueue_head(&queue);
	while ((atomic_read(&is_dead) == 0) && count < 3) {
		//lower semaphore for the thread
		atomic_set(&kill_thread, 1);
//        wake_up_interruptible(&my_wait);
		wake_up_process(spbmthread);
		interruptible_sleep_on_timeout(&queue, 50);

		//sleep for a bit and then check on the thread again
		count++;

	}
}

void sendBeacon(struct spbm_event *evt)
{
	sendPacket(0, 0, SPBM_TYPE_BEACON, 0);
	event_set_time(evt, BEACONTIME);
	event_add(evt);
}

void sendPacket(char *payload, int len, int type, char *dest)
{
	struct sk_buff *skb;
	struct net_device *dev;
	struct spbmbeacon *beacon;
	struct ethhdr *eth;
	if (getXPos() == -1) {
		printk("Won't send packet because position is unknown\n");
		return;
	}

	dev = getDev();

	if (!dev) {
		printk("couldn't find %s", DEVNAME);
		return;
	}
//      printk("found device %s\n",dev->name);

	skb = alloc_skb(sizeof(struct spbmbeacon) +
			+dev->hard_header_len + 15 + len, GFP_ATOMIC);
	if (!skb) {
		printk("SPBM: couldn't alloc skb in sendPacket()\n");
		return;
	}

	skb_reserve(skb, (dev->hard_header_len + 15) & ~15);
	skb->nh.raw = skb->data;
	beacon =
	    (struct spbmbeacon *) skb_put(skb, sizeof(struct spbmbeacon));
	if (!beacon) {
		printk("skb_put in sendBeacon() failed\n");
		kfree_skb(skb);
		return;
	}
	beacon->type = type;
	beacon->pos.x = htons(getXPos());
	beacon->pos.y = htons(getYPos());

	if (payload) {
		char *pl = skb_put(skb, len);
		memcpy(pl, payload, len);
	}

	skb->dev = dev;
	skb->protocol = __constant_htons(PROT_SPBM);

	eth = (struct ethhdr *) skb_push(skb, sizeof(struct ethhdr));
	eth->h_proto = htons(PROT_SPBM);
	memcpy(eth->h_source, dev->dev_addr, dev->addr_len);
	if (!dest)
		memcpy(eth->h_dest, dev->broadcast, dev->addr_len);
	else
		memcpy(eth->h_dest, dest, dev->addr_len);
	dev_queue_xmit(skb);
//      printk("spbm: packet sent (type: %d len: %d)!\n",type,len);
// XXX netfilter
//      NF_HOOK(NF_IP, NF_IP_LOCAL_OUT, skb, NULL, dev, dev_queue_xmit);
	dev_put(dev);
	return;
}

struct spbm_event *find_old_event()
{
	struct list_head *n;
	struct timeval tv;
	unsigned long flags;
	do_gettimeofday(&tv);
	normalizetv(&tv);
	read_lock_irqsave(&event_lock, flags);
	list_for_each(n, &events) {
		if (((struct spbm_event *) n)->t.tv_sec < tv.tv_sec ||
		    (((struct spbm_event *) n)->t.tv_sec == tv.tv_sec &&
		     ((struct spbm_event *) n)->t.tv_usec < tv.tv_usec)) {
			read_unlock_irqrestore(&event_lock, flags);
			return (struct spbm_event *) n;
		}
	}
	read_unlock_irqrestore(&event_lock, flags);
	return 0;
}

struct spbm_event neighbourevt;

void cleanNeighbours(struct spbm_event *evt);

void mythread()
{
	struct timeval tv;
	struct timeval tv2;
	int timeout;
	unsigned long flags;
	struct spbm_event *evt;

	//Initalize the variables
	init_waitqueue_head(&my_wait);
	atomic_set(&kill_thread, 0);
	atomic_set(&is_dead, 0);

	//Name our thread
	lock_kernel();
	sprintf(current->comm, "spbm");
	exit_mm(current);
	unlock_kernel();

	spbmthread = current;

	printk("thread started \n");

	beaconevt.action = sendBeacon;
	event_set_time(&beaconevt, 5000);
	event_add(&beaconevt);

	neighbourevt.action = cleanNeighbours;
	event_set_time(&neighbourevt, 5000);	// first table clean in 5sec
	event_add(&neighbourevt);

	initupdateevents();

	while (atomic_read(&kill_thread) == 0) {
		do_gettimeofday(&tv);
		do_gettimeofday(&tv2);
		normalizetv(&tv);
		normalizetv(&tv2);

		while (evt = find_old_event()) {
			write_lock_irqsave(&event_lock, flags);
			// r/w lock ordering save because nowhere else are events deleted
			list_del((struct list_head *) evt);
			write_unlock_irqrestore(&event_lock, flags);
//              printk("calling an events action\n");
			evt->action(evt);
		}
		if (list_empty(&events)) {
			timeout = 10000;
			printk("No events!\n");
		} else {
			struct list_head *n;
			int cnt = 0;
			tv.tv_sec += 100;
			read_lock_irqsave(&event_lock, flags);
			list_for_each(n, &events) {
				if (((struct spbm_event *) n)->t.tv_sec ==
				    tv.tv_sec) {
					if (((struct spbm_event *) n)->t.
					    tv_usec <= tv.tv_usec)
						tv.tv_usec =
						    (((struct spbm_event *)
						      n)->t.tv_usec);
				}
				if (((struct spbm_event *) n)->t.tv_sec <
				    tv.tv_sec) {
					tv.tv_usec =
					    (((struct spbm_event *) n)->t.
					     tv_usec);
					tv.tv_sec =
					    (((struct spbm_event *) n)->t.
					     tv_sec);
				}
				++cnt;
			}
			read_unlock_irqrestore(&event_lock, flags);
			// dependent on number of context switches per second
			timeout =
			    (tv.tv_sec - tv2.tv_sec) * 100 + (tv.tv_usec -
							      tv2.
							      tv_usec) /
			    10000;
//              printk("new timeout %d (%d events)\n",timeout,cnt);
		}
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(timeout > 0 ? timeout : 1);
//      interruptible_sleep_on_timeout(&my_wait,timeout>0 ? timeout : 1);
	}
	atomic_set(&is_dead, 1);
	printk("thread stopped");

}

void startme()
{
	pid = kernel_thread((void *) &mythread, NULL, 0);
}
